"""Useful functions for the project."""
import numpy as np
import argparse
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.interpolate import PchipInterpolator
import warnings


def amplitude_using_all_modes(mode_dict):
    """Get the amplitude using all the available modes.

    Parameters:
    ----------
    mode_dict:
        Dictionary containing waveform modes.

    Returns:
        Square root of the qudrature sum of the amplitudes of all the
        available modes in mode_dict.
    """
    amp = 0
    for mode in mode_dict.keys():
        amp += np.abs(mode_dict[mode])**2
    return np.sqrt(amp)


def peak_time_via_quadratic_fit(t, func):
    """
    Find the peak time of a function quadratically.

    Fits the function to a quadratic over the 5 points closest to the argmax
    func.

    Parameters:
    -----------
    t:
        An array of times.
    func:
        Array of function values.

    Returns:
    --------
    tpeak:
        Time at peak of the function func.
    fpeak:
        Value of function func at tpeak.
    """
    # Find the time closest to the peak, making sure we have room on either
    # side
    index = np.argmax(func)
    index = max(2, min(len(t) - 3, index))

    # Do a quadratic fit to 5 points,
    # subtracting t[index] to make the matrix inversion nice
    testTimes = t[index-2:index+3] - t[index]
    testFuncs = func[index-2:index+3]
    xVecs = np.array([np.ones(5), testTimes, testTimes**2.])
    invMat = np.linalg.inv(np.array([[v1.dot(v2) for v1 in xVecs]
                                     for v2 in xVecs]))

    yVec = np.array([testFuncs.dot(v1) for v1 in xVecs])
    coefs = np.array([yVec.dot(v1) for v1 in invMat])
    return t[index] - coefs[1]/(2.*coefs[2]), (coefs[0]
                                               - coefs[1]**2./4/coefs[2])


def check_kwargs_and_set_defaults(user_kwargs=None,
                                  default_kwargs=None,
                                  name="user given kwargs",
                                  location=None):
    """Sanity check user given dicionary of kwargs and set default values.

    Parameters:
    ----------
    user_kwargs:
        Dictionary of kwargs by user.
    default_kwargs:
        Dictionary of default kwargs.
    name:
        string to represent the dictionary
    location:
        string pointing to where the defaults are defined

    Returns:
    --------
    updated user_kwargs
    """
    # make user_kwargs iterable
    if user_kwargs is None:
        user_kwargs = {}

    for kw in user_kwargs.keys():
        if kw not in default_kwargs:
            raise ValueError(f"Invalid key {kw} in {name}."
                             " Should be one of "
                             f"{list(default_kwargs.keys())}\n"
                             f"To add a new keyword, please modify {location}")

    for kw in default_kwargs.keys():
        if kw not in user_kwargs:
            user_kwargs[kw] = default_kwargs[kw]

    return user_kwargs


def raise_exception_if_none(kwargs, keys_to_check, name, location):
    """Raise exception if any key from `keys_to_check` has value None."""
    for kw in keys_to_check:
        if kwargs[kw] is None:
            raise Exception(f"kw {kw} for {name} can not be None."
                            f" Check documentation of {location} for more"
                            " details.")


class SmartFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """Smart Formatter for argparse helper strings.

    Stolen from https://stackoverflow.com/questions/3853722/how-to-insert-newlines-on-argparse-help-text.
    """

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)


def time_deriv_4thOrder(y, dt):
    """Fourth order accurate time derivative.

    Assuming constant time step.
    Tested for convergence up to 1e-12 level.

    Parameters:
    -----------
    y:
        1d array to take time derivative of.
    dt:
        Time step.

    Returns:
    --------
    dydt:
        Fourth order time derivative of y.
    """
    # Use a 5 point stencil
    res = 0*y
    # First do the interior
    res[2:-2] = (y[:-4] - 8*y[1:-3] + 8*y[3:-1] - y[4:])/12.

    # Next do the edges
    res[1] = y[:5].dot(np.array([-3, -10, 18, -6, 1]) / 12.)
    res[0] = y[:5].dot(np.array([-25, 48, -36, 16, -3]) / 12.)
    res[-2] = y[-5:].dot(np.array([-1, 6, -18, 10, 3]) / 12.)
    res[-1] = y[-5:].dot(np.array([3, -16, 36, -48, 25]) / 12.)
    dydt = res / dt
    return dydt


def get_default_spline_kwargs():
    """Defaults for spline settings."""
    default_spline_kwargs = {
        "w": None,
        "bbox": [None, None],
        "k": 3,
        "ext": 2,
        "check_finite": False}
    return default_spline_kwargs


def interpolate(newX,
                oldX,
                oldY,
                allowExtrapolation=False,
                interpolator="spline",
                spline_kwargs=None,
                check_kwargs=True):
    """Interpolate.

    Parameters:
    -----------
    newX:
        Points where interpolant is to be evaluated.
    oldX:
        1d array of monotonically increasing real values.
    oldY:
        1d array of monotonically increasing real values.
    allowExtrapolation:
        Bool. If True returns extrapolated values. Default is False.
        If False, an exception is raised if trying to extrapolate.
    interpolator:
        String to choose an interpolator to interpolate oldX, oldY.
        Could be one of the following:
        "spline": Uses scipy.interpolate.InterpolatedUnivariateSpline.
        "monotonic_spline":  Uses scipy.interpolate.PchipInterpolator.
        Default is "spline".
    spline_kwargs:
        See under get_interpolant.
    check_kwargs:
        Check spline_kwargs if check_kwargs is True. Default is True.

    Returns:
    --------
    newY:
        Intepolated values at newX.
    """
    if len(oldY) != len(oldX):
        raise Exception("Lengths dont match.")

    if not allowExtrapolation:
        if np.min(newX) < np.min(oldX) - 1e-10 \
           or np.max(newX) > np.max(oldX) + 1e-10:
            print(f"Min of newX = {np.min(newX)}, "
                  f"Min of oldX = {np.min(oldX)}")
            print(f"Max of newX = {np.max(newX)}, "
                  f"Max of oldX = {np.max(oldX)}")
            print("newX has vlaues below oldX? "
                  f"{np.min(newX) < np.min(oldX)}")
            print("newX has values above oldX? "
                  f"{np.max(newX) > np.max(oldX)}")
            raise Exception("Trying to extrapolate, "
                            "but allowExtrapolation=False")
    newY = get_interpolant(oldX, oldY, allowExtrapolation, interpolator,
                           spline_kwargs, check_kwargs)(newX)
    return newY


def get_interpolant(oldX,
                    oldY,
                    allowExtrapolation=False,
                    interpolator="spline",
                    spline_kwargs=None,
                    check_kwargs=True):
    """Create Interpolant.

    Parameters:
    -----------
    oldX:
        1d array of monotonically increasing real values.
    oldY:
        1d array of monotonically increasing real values.
    allowExtrapolation:
        Bool. If True returns extrapolated values. Default is False.
        If False, an exception is raised if trying to extrapolate.
    interpolator:
        String to choose an interpolator to interpolate oldX, oldY.
        Could be one of the following:
        "spline": Uses scipy.interpolate.InterpolatedUnivariateSpline.
        "monotonic_spline":  Uses scipy.interpolate.PchipInterpolator.
        Default is "spline".
    spline_kwargs:
        Dictionary of kwargs to be provided to the interpolator if
        "interpolator"="spline". The allowed kwargs are the same as that of the
        spline function scipy.interpolate.InterpolatedUnivariateSpline and the
        defaults are set using gw_eccentricity.utils.get_default_spline_kwargs.
        Since we use allowExtraplotion arg separately, value of "ext" in
        extra_kwargs will be overridden by allowExtrapolation.
    check_kwargs:
        Check spline_kwargs if check_kwargs is True. Default is True.

    Returns:
    --------
    Intepolatnt.
    """
    if not np.all(np.diff(oldX) > 0):
        raise Exception("oldX must have increasing values")

    if interpolator == "spline":
        if check_kwargs:
            spline_kwargs = check_kwargs_and_set_defaults(
                spline_kwargs, get_default_spline_kwargs(), "spline kwargs",
                "utils.get_default_spline_kwargs")
        # check that num of data points > order of spline
        if len(oldX) >= 2:
            if len(oldX) <= spline_kwargs["k"]:
                warnings.warn(f"No of data points is {len(oldX)} but "
                              f"spline order k = {spline_kwargs['k']}. "
                              f"Decreasing k to {len(oldX) - 1}.")
                # make a copy so that the original spline_kwargs remains
                # unmodified.
                kwargs = spline_kwargs.copy()
                kwargs["k"] = len(oldX) - 1
            else:
                kwargs = spline_kwargs
        else:
            raise Exception("Number of data points is {len(oldX)}."
                            " Cannot build an interpolant.")
        # If allowExtrapolation is True but ext=2 then raise a warning
        # and override ext to 0
        if allowExtrapolation and kwargs["ext"] == 2:
            # ext = 0, returns extraploted values
            kwargs["ext"] = 0
        # If allowExtraplotion is False but ext != 2 then raise a warning
        # and override ext to 2
        if not allowExtrapolation and kwargs["ext"] != 2:
            # ext = 2, raises exception if extrapolation is attempted.
            kwargs["ext"] = 2
        interpolant = InterpolatedUnivariateSpline(oldX, oldY, **kwargs)
    elif interpolator == "monotonic_spline":
        if spline_kwargs is not None:
            warnings.warn(f"Interpolator is {interpolator} but spline_kwargs "
                          "are passed. spline_kwargs will be ignored.")
        interpolant = PchipInterpolator(oldX, oldY,
                                        extrapolate=allowExtrapolation)
    else:
        raise ValueError(f"Unknown interpolator {interpolator}. Must be one"
                         " of ['spline', 'monotonic_spline']")
    return interpolant


def debug_message(message, debug_level, important=True,
                  point_to_verbose_output=False):
    """Show message based on debug_level.

    parameters:
    -----------
    message: str
        Message to display.

    debug_level: int
        Indicator for level of debug message. Based on it, one of the
        following actions if performed:
        -1: No action is performed and hence no message is displayed.
        0: Warning is issued with the input message only if important=True
        1: Warning is issued with the input message.
        2: Exception is raised with the input message.

    important: bool
        Only if True, the message gets printed when debug_level=0. For
        other debug_levels, this does nothing.
        Default is True.

    point_to_verbose_output: bool
        When True, if debug_level is 0 and important is True, points to
        debug_level = 1 for more verbose output.
        Default is False
    """
    debug_levels = [-1, 0, 1, 2]
    if debug_level not in debug_levels:
        raise ValueError(
            f"Unknown debug_level {debug_level}. Should one "
            f"of {debug_levels}. See "
            "`gw_eccentricity.utils.debug_message` for action"
            "performed with each debug level.")
    if debug_level == -1:
        # Do nothing
        return
    if (debug_level == 0 and important) or debug_level == 1:
        if debug_level == 0 and point_to_verbose_output:
            message += "\nFor more verbose output use `debug_level=1`."
        # Issue warning. Use stacklevel=2 to point to actual line number
        # causing this warning instead of pointing to here.
        warnings.warn(message, stacklevel=2)
    if debug_level == 2:
        # raise Exception
        raise Exception(message)



"""Regression test."""
import numpy as np
import json
import gw_eccentricity
from gw_eccentricity import load_data
from gw_eccentricity import measure_eccentricity

# locally it passs without problem but on github action, we need to set this tolerance
atol = 1e-9

def test_regression():
    """Regression test using all methods."""
    # List of all available methods
    available_methods = gw_eccentricity.get_available_methods()
    for method in available_methods:
        # Load regression data
        regression_data_file = f"test/regression_data/{method}_regression_data.json"
        # Load the regression data
        fl = open(regression_data_file, "r")
        regression_data = json.load(fl)
        fl.close()
        # waveform kwargs
        lal_kwargs = regression_data["waveform_kwargs"]
        # load waveform data
        dataDict = load_data.load_waveform(**lal_kwargs)
        extra_kwargs = regression_data["extra_kwargs"]

        # Try evaluating at times where regression data are saved
        regression_data_at_tref = regression_data["tref"]
        tref_in = regression_data_at_tref["time"]
        gwecc_dict = measure_eccentricity(
            tref_in=tref_in,
            method=method,
            dataDict=dataDict,
            extra_kwargs=extra_kwargs)
        tref_out = gwecc_dict["tref_out"]
        ecc_ref = gwecc_dict["eccentricity"]
        meanano_ref = gwecc_dict["mean_anomaly"]
        # Compare the measured data with the saved data
        np.testing.assert_allclose(
            ecc_ref, regression_data_at_tref["eccentricity"],
            atol=atol,
            err_msg="measured and saved eccentricity at saved times do not match.")
        np.testing.assert_allclose(
            meanano_ref, regression_data_at_tref["mean_anomaly"],
            atol=atol,
            err_msg="measured and saved mean anomaly at saved times do not match.")

        # Try evaluating at frequencies where regression data are saved
        regression_data_at_fref = regression_data["fref"]
        fref_in = regression_data_at_fref["frequency"]
        gwecc_dict = measure_eccentricity(
            fref_in=fref_in,
            method=method,
            dataDict=dataDict,
            extra_kwargs=extra_kwargs)
        fref_out = gwecc_dict["fref_out"]
        ecc_ref = gwecc_dict["eccentricity"]
        meanano_ref = gwecc_dict["mean_anomaly"]
        # Compare the measured data with the saved data
        np.testing.assert_allclose(
            ecc_ref, regression_data_at_fref["eccentricity"],
            atol=atol,
            err_msg="measured and saved eccentricity at saved frequencies do not match.")
        np.testing.assert_allclose(
            meanano_ref, regression_data_at_fref["mean_anomaly"],
            atol=atol,
            err_msg="measured and saved mean anomaly at saved frequencies do not match.")



"""Group NR waveforms in precessing and non-precessing directory.

April 13, 2022
Md Arif Shaikh
"""
import h5py
import numpy as np
import argparse
import os
import warnings
import glob
from tqdm import tqdm


parser = argparse.ArgumentParser()
parser.add_argument(
    "--dest_dir",
    type=str,
    help=("Main destination directory where the NR files would be stored after"
          " categorizing. This is where the sub directories would live."))
parser.add_argument(
    "--nr_dir",
    type=str,
    help=("NR directory where the NR files are located before"
          " categorizing. If this is provided then all files"
          "under this directory would be worked on. If you want"
          " to work on a single file, use --nr_file key."))
parser.add_argument(
    "--nr_file",
    type=str,
    help=("NR file to be categorized."))
parser.add_argument(
    "--catalog",
    type=str,
    required=True,
    help=("Name of the waveform catalog. Choose the existing names"
          " from get_standard_catalog_names if possible to avoid duplicating"
          " directories."))

args = parser.parse_args()


def get_standard_catalog_names():
    """Get the list of standard catalog names."""
    dict_of_catalog_names = {
        "SXS": "Numerical Relativity waveforms from SXS collaboration",
        "RIT": "Numerical Relativity waveforms from RIT",
        "MAYA": "Numerical Relativity waveforms from MAYA",
        "ET": "Numerical Relativity waveforms from Einstein Toolkitt",
        "SEOB": "SEOB waveforms from Toni's model."}
    return dict_of_catalog_names


# warnings if catalog name is non standard
if args.catalog not in get_standard_catalog_names():
    warnings.warn(f"Catalog name {args.catalog} is not one of the standard"
                  f" catalog names {get_standard_catalog_names().keys()}")


def check_precessing(waveform_file):
    """Check if the given waveform is precessing."""
    f = h5py.File(waveform_file, "r")
    s1x = f.attrs["spin1x"]
    s1y = f.attrs["spin1y"]
    s1z = f.attrs["spin1z"]
    s2x = f.attrs["spin2x"]
    s2y = f.attrs["spin2y"]
    s2z = f.attrs["spin2z"]
    chi1 = [s1x, s1y, s1z]
    chi2 = [s2x, s2y, s2z]
    f.close()
    return (np.linalg.norm(chi1[:2]) > 1e-3
            or np.linalg.norm(chi2[:2]) > 1e-3)


def create_dir(dir):
    """Create a dir if it does not exist."""
    if not os.path.exists(dir):
        os.system(f"mkdir -p {dir}")


def move_file(waveform_file, dest_dir, catalog):
    """Move a file to correc sub directory."""
    sub_dir = "" if check_precessing(waveform_file) else "Non-"
    dest_dir = f"{dest_dir}/{sub_dir}Precessing/{catalog}"
    create_dir(dest_dir)
    fileName = os.path.basename(waveform_file)  # won't work on windows
    dest_path = f"{dest_dir}/{fileName}"
    os.system(f"cp {waveform_file} {dest_path}")


# move nr files to precessing/Non-precessing directories
if args.nr_file:
    move_file(args.nr_file, args.dest_dir, args.catalog)
else:
    if not args.nr_dir:
        raise Exception("Must provide nr_dir if nr_file is not given.")
    nr_files = glob.glob(f"{args.nr_dir}/*.h5")
    for fileName in tqdm(nr_files):
        move_file(fileName, args.dest_dir, args.catalog)



"""gw_eccentricity wrapper to call different eccDefinition methods.

Measure eccentricity and mean anomaly from gravitational waves.
See our paper https://arxiv.org/abs/xxxx.xxxx and
https://pypi.org/project/gw_eccentricity for more details.
FIXME ARIF: Add arxiv link when available.
"""
__copyright__ = "Copyright (C) 2023 Md Arif Shaikh, Vijay Varma, Harald Pfeiffer"
__email__ = "arifshaikh.astro@gmail.com, vijay.varma392@gmail.com"
__status__ = "testing"
__author__ = "Md Arif Shaikh, Vijay Varma, Harald Pfeiffer"
__version__ = "0.0.3"
__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from .eccDefinitionUsingAmplitude import eccDefinitionUsingAmplitude
from .eccDefinitionUsingFrequency import eccDefinitionUsingFrequency
from .eccDefinitionUsingAmplitudeFits import eccDefinitionUsingAmplitudeFits
from .eccDefinitionUsingFrequencyFits import eccDefinitionUsingFrequencyFits
from .eccDefinitionUsingResidualAmplitude import eccDefinitionUsingResidualAmplitude
from .eccDefinitionUsingResidualFrequency import eccDefinitionUsingResidualFrequency


def get_available_methods(return_dict=False):
    """Get all available eccDefinition methods.

    If return_dict is True, returns a dictionary of methods.
    Else, just returns a list of method names.
    """
    methods = {
        "Amplitude": eccDefinitionUsingAmplitude,
        "Frequency": eccDefinitionUsingFrequency,
        "ResidualAmplitude": eccDefinitionUsingResidualAmplitude,
        "ResidualFrequency": eccDefinitionUsingResidualFrequency,
        "AmplitudeFits": eccDefinitionUsingAmplitudeFits,
        "FrequencyFits": eccDefinitionUsingFrequencyFits
    }

    if return_dict:
        return methods
    else:
        return list(methods.keys())


def measure_eccentricity(tref_in=None,
                         fref_in=None,
                         method="Amplitude",
                         dataDict=None,
                         num_orbits_to_exclude_before_merger=2,
                         extra_kwargs=None):
    """Measure eccentricity and mean anomaly from a gravitational waveform.

    Eccentricity is measured using the GW frequency omega22(t) =
    dphi22(t)/dt, where phi22(t) is the phase of the (2, 2) waveform
    mode. We currently only allow time-domain, nonprecessing waveforms. We
    evaluate omega22(t) at pericenter times, t_pericenters, and build a
    spline interpolant omega22_pericenters(t) using those data
    points. Similarly, we build omega22_apocenters(t) using omega22(t) at
    the apocenter times, t_apocenters.

    Using omega22_pericenters(t) and omega22_apocenters(t), we first
    compute e_omega22(t), as described in Eq.(4) of arxiv:xxxx.xxxx. We
    then use e_omega22(t) to compute the eccentricity egw(t) using Eq.(8)
    of arxiv:xxxx.xxxx. Mean anomaly is defined using t_pericenters, as
    described in Eq.(10) of arxiv:xxxx.xxxx.

    FIXME ARIF: In the above text, fill in arxiv number when
    available.

    To find t_pericenters/t_apocenters, one can look for extrema in different
    waveform data, like omega22(t) or Amp22(t), the amplitude of the (2, 2)
    mode. Pericenters correspond to the local maxima, while apocenters
    correspond to the local minima in the data. The method option
    (described below) lets the user pick which waveform data to use to find
    t_pericenters/t_apocenters.

    Parameters:
    ----------
    tref_in:
        Input reference time at which to measure eccentricity and mean anomaly.
        Can be a single float or an array.

    fref_in:
        Input reference GW frequency at which to measure the eccentricity and
        mean anomaly. Can be a single float or an array. Only one of
        tref_in/fref_in should be provided.

        Given an fref_in, we find the corresponding tref_in such that
        omega22_average(tref_in) = 2 * pi * fref_in. Here, omega22_average(t)
        is a monotonically increasing average frequency obtained from the
        instantaneous omega22(t). omega22_average(t) defaults to the orbit
        averaged omega22, but other options are available (see
        omega22_averaging_method below).

        Eccentricity and mean anomaly measurements are returned on a subset of
        tref_in/fref_in, called tref_out/fref_out, which are described below.
        If dataDict is provided in dimensionless units, tref_in should be in
        units of M and fref_in should be in units of cycles/M. If dataDict is
        provided in MKS units, t_ref should be in seconds and fref_in should be
        in Hz.

    method: str
        Which waveform data to use for finding extrema. Options are:
        - "Amplitude": Finds extrema of Amp22(t).
        - "Frequency": Finds extrema of omega22(t).
        - "ResidualAmplitude": Finds extrema of resAmp22(t), the residual
          amplitude, obtained by subtracting the Amp22(t) of the quasicircular
          counterpart from the Amp22(t) of the eccentric waveform. The
          quasicircular counterpart is described in the documentation of
          dataDict below.
        - "ResidualFrequency": Finds extrema of resomega22(t), the residual
          frequency, obtained by subtracting the omega22(t) of the
          quasicircular counterpart from the omega22(t) of the eccentric
          waveform.
        - "AmplitudeFits": Uses Amp22(t) and iteratively subtracts a
          PN-inspired fit of the extrema of Amp22(t) from it, and finds extrema
          of the residual.
        - "FrequencyFits": Uses omega22(t) and iteratively subtracts a
          PN-inspired fit of the extrema of omega22(t) from it, and finds
          extrema of the residual.
        Default is "Amplitude".
        The available list of methods can be also obtained from
        gw_eccentricity.get_available_methods().
        Detailed description of these methods can be found in Sec. III of
        arxiv.xxxx.xxxx.

        The Amplitude and Frequency methods can struggle for very small
        eccentricities, especially near the merger, as the
        secular amplitude/frequency growth dominates the modulations due to
        eccentricity, making extrema finding difficult.

        The ResidualAmplitude/ResidualFrequency/AmplitudeFits/FrequencyFits
        methods avoid this limitation by removing the secular growth before
        finding extrema. However, methods that use the frequency for finding
        extrema (Frequency/ResidualFrequency/FrequencyFits) can be more
        sensitive to junk radiation in NR data.

        Therefore, the recommended methods are
        ResidualAmplitude/AmplitudeFits/Amplitude

    dataDict:
        Dictionary containing waveform modes dict, time etc. Should follow the
        format:
        dataDict = {"t": time,
                    "hlm": modeDict,
                    "t_zeroecc": time,
                    "hlm_zeroecc": modeDict,
                   },
        "t" and "hlm" are mandatory. "t_zeroecc" and "hlm_zeroecc" are only
        required for ResidualAmplitude and ResidualFrequency methods, but if
        provided, they are used for additional diagnostic plots, which can be
        helpful for all methods. Any other keys in dataDict will be ignored,
        with a warning.

        The recognized keys are:
        - "t": 1d array of times.
            - Should be uniformly sampled, with a small enough time step so
              that omega22(t) can be accurately computed. We use a 4th-order
              finite difference scheme. In dimensionless units, we recommend a
              time step of dtM = 0.1M to be conservative, but you may be able
              to get away with larger time steps like dtM = 1M. The
              corresponding time step in seconds would be
              dtM * M * lal.MTSUN_SI, where M is the total mass in Solar
              masses.
            - We do not require the waveform peak amplitude to occur at any
              specific time, but tref_in should follow the same convention for
              peak time as "t".
        - "hlm": Dictionary of waveform modes associated with "t".
            - Should have the format:
                modeDict = {(l1, m1): h_{l1, m1},
                            (l2, m2): h_{l2, m2},
                            ...
                           },
                where h_{l, m} is a 1d complex array representing the (l, m)
                waveform mode. Should contain at least the (2, 2) mode, but
                more modes can be included, as indicated by the ellipsis '...'
                above.
        - "t_zeroecc" and "hlm_zeroecc":
            - Same as above, but for the quasicircular counterpart to the
              eccentric waveform. The quasicircular counterpart can be obtained
              by evaluating a waveform model by keeping the rest of the binary
              parameters fixed (same as the ones used to generate "hlm") but
              setting the eccentricity to zero. For NR, if such a quasicircular
              counterpart is not available, we recommend using quasicircular
              models like NRHybSur3dq8 or IMRPhenomT, depending on the mass
              ratio and spins.
            - "t_zeroecc" should be uniformly spaced, but does not have to
              follow the same time step as that of "t", as long as the step
              size is small enough to compute the frequency. Similarly, peak
              time does not have to match that of "t".
            - We require that "hlm_zeroecc" be at least as long as "hlm" so
              that residual amplitude/frequency can be computed.

    num_orbits_to_exclude_before_merger:
        Can be None or a non negative number.
        If None, the full waveform data (even post-merger) is used for
        finding extrema, but this might cause interpolation issues.
        For a non negative num_orbits_to_exclude_before_merger, that
        many orbits before the merger are excluded when finding extrema.
        If your waveform does not have a merger (e.g. PN/EMRI), use
        num_orbits_to_exclude_before_merger = None.

        The default value is chosen via an investigation on a set of NR
        waveforms. See the following wiki page for more details,
        https://github.com/vijayvarma392/gw_eccentricity/wiki/NR-investigation-to-set-default-number-of-orbits-to-exclude-before-merger
        Default: 2.

    extra_kwargs: A dict of any extra kwargs to be passed. Allowed kwargs are:
        spline_kwargs:
            Dictionary of arguments to be passed to the spline interpolation
            routine (scipy.interpolate.InterpolatedUnivariateSpline) used to
            compute quantities like omega22_pericenters(t) and
            omega22_apocenters(t).
            Defaults are set using utils.get_default_spline_kwargs

        extrema_finding_kwargs:
            Dictionary of arguments to be passed to the extrema finder,
            scipy.signal.find_peaks.
            The Defaults are the same as those of scipy.signal.find_peaks,
            except for the "width", which sets the minimum allowed "full width
            at half maximum" for the extrema. Setting this can help avoid
            false extrema in noisy data (for example, due to junk radiation in
            NR). The default for "width" is set using phi22(t) near the
            merger. Starting from 4 cycles of the (2, 2) mode before the
            merger, we find the number of time steps taken to cover 2 cycles,
            let's call this "the gap". Note that 2 cycles of the (2, 2) mode
            are approximately one orbit, so this allows us to approximate the
            smallest gap between two pericenters/apocenters. However, to be
            conservative, we divide this gap by 4 and set it as the width
            parameter for find_peaks. See
            eccDefinition.get_width_for_peak_finder_from_phase22 for more
            details.

        debug_level: int
            Debug settings for warnings/errors:
            -1: All warnings are suppressed. NOTE: Use at your own risk!
            0: Only important warnings are issued.
            1: All warnings are issued. Use when investigating.
            2: All warnings become exceptions.
            Default: 0.

        debug_plots: bool
            If True, diagnostic plots are generated. This can be
            computationally expensive and should only be used when
            debugging. When True, look for figures saved as
            `gwecc_{method}_*.pdf`.
            Default is False.

        omega22_averaging_method:
            Options for obtaining omega22_average(t) from the instantaneous
            omega22(t).
            - "orbit_averaged_omega22": First, orbit averages are obtained at
              each pericenter by averaging omega22(t) over the time from the
              current pericenter to the next one. This average value is
              associated with the time at midpoint between the current and the
              next pericenter. Similarly, orbit averages are computed at
              apocenters.  Finally, a spline interpolant is constructed between
              all of these orbit averages at extrema locations. However, the
              final time over which the spline is constructed is constrained to
              be between tmin_for_fref and tmax_for_fref which are close to
              tmin and tmax, respectively. See eccDefinition.get_fref_bounds()
              for details.
            - "mean_of_extrema_interpolants":
              The mean of omega22_pericenters(t) and omega22_apocenters(t) is
              used as a proxy for the average frequency.
            - "omega22_zeroecc": omega22(t) of the quasicircular counterpart
              is used as a proxy for the average frequency. This can only be
              used if "t_zeroecc" and "hlm_zeroecc" are provided in dataDict.
            See Sec. IID of arxiv.xxxx.xxxx for a more detailed description of
            "omega22_average".
            Default is "orbit_averaged_omega22".

        treat_mid_points_between_pericenters_as_apocenters:
            If True, instead of trying to find apocenter locations by looking
            for local minima in the data, we simply find the midpoints between
            pericenter locations and treat them as apocenters. This is helpful
            for eccentricities ~1 where pericenters are easy to find but
            apocenters are not.
            Default: False.

        kwargs_for_fits_methods:
            Extra kwargs to be passed to FrequencyFits and AmplitudeFits
            methods. See
            eccDefinitionUsingFrequencyFits.get_default_kwargs_for_fits_methods
            for allowed keys.

    Returns:
    --------
    A dictionary containing the following keys
    tref_out:
        tref_out is the output reference time at which eccentricity and mean
        anomaly are measured.
        tref_out is included in the returned dictionary only when tref_in is
        provided.
        Units of tref_out are the same as that of tref_in.

        tref_out is set as
        tref_out = tref_in[tref_in >= tmin & tref_in <= tmax],
        where tmax = min(t_pericenters[-1], t_apocenters[-1]) and
              tmin = max(t_pericenters[0], t_apocenters[0]),
        As eccentricity measurement relies on the interpolants
        omega22_pericenters(t) and omega22_apocenters(t), the above cutoffs
        ensure that we only compute the eccentricity where both
        omega22_pericenters(t) and omega22_apocenters(t) are within their
        bounds.

    fref_out:
        fref_out is the output reference frequency at which eccentricity and
        mean anomaly are measured.
        fref_out is included in the returned dictionary only when fref_in is
        provided.
        Units of fref_out are the same as that of fref_in.

        fref_out is set as
        fref_out = fref_in[fref_in >= fref_min && fref_in <= fref_max],
        where fref_min/fref_max are minimum/maximum allowed reference
        frequency, with fref_min = omega22_average(tmin_for_fref)/2/pi
        and fref_max = omega22_average(tmax_for_fref)/2/pi.
        tmin_for_fref/tmax_for_fref are close to tmin/tmax, see
        eccDefinition.get_fref_bounds() for details.

    eccentricity:
        Measured eccentricity at tref_out/fref_out. Same type as
        tref_out/fref_out.

    mean_anomaly:
        Measured mean anomaly at tref_out/fref_out. Same type as
        tref_out/fref_out.

    gwecc_object:
        eccDefinition object used to compute eccentricity. This can be used to
        make diagnostic plots, call member functions to do further
        investigation or access variables computed internally for measuring
        eccentricity and mean anomaly.
    """
    available_methods = get_available_methods(return_dict=True)

    if method in available_methods:
        gwecc_object = available_methods[method](
            dataDict, num_orbits_to_exclude_before_merger, extra_kwargs)
        return_dict = gwecc_object.measure_ecc(
            tref_in=tref_in, fref_in=fref_in)
        return_dict.update({"gwecc_object": gwecc_object})
        return return_dict
    else:
        raise Exception(f"Invalid method {method}, has to be one of"
                        f" {list(available_methods.keys())}")



"""
Find pericenters and apocenters using Amplitude.

Part of Eccentricity Definition project.
Md Arif Shaikh, Mar 28, 2022
"""
from .eccDefinition import eccDefinition
from scipy.signal import find_peaks


class eccDefinitionUsingAmplitude(eccDefinition):
    """Define eccentricity by finding extrema location from amplitude."""

    def __init__(self, *args, **kwargs):
        """Init for eccDefinitionUsingAmplitude class.

        parameters:
        ----------
        dataDict: Dictionary containing the waveform data.
        """
        super().__init__(*args, **kwargs)
        self.data_for_finding_extrema = self.get_data_for_finding_extrema()
        self.label_for_data_for_finding_extrema = r"$A_{22}$"
        self.method = "Amplitude"

    def get_data_for_finding_extrema(self):
        """Get data to be used for finding extrema location.

        In the derived classes, one need to override this function
        to return the appropriate data that is to be used. For example,
        in residual amplitude method, this function would return
        residual amp22, whereas for frequency method, it would
        return omega22 and so on.
        """
        return self.amp22

    def find_extrema(self, extrema_type="pericenters"):
        """Find the extrema in the amp22.

        parameters:
        -----------
        extrema_type:
            Either "pericenters" or "apocenters".

        returns:
        ------
        array of positions of extrema.
        """
        if extrema_type == "pericenters":
            sign = 1
        elif extrema_type == "apocenters":
            sign = - 1
        else:
            raise Exception("`extrema_type` must be either 'pericenters'"
                            " or 'apocenters'")

        return find_peaks(
            sign * self.data_for_finding_extrema,
            **self.extrema_finding_kwargs)[0]



"""setup for gw_eccentricity."""
import setuptools

with open("README.md", "rb") as fh:
    long_description = fh.read().decode("UTF-8")


# Extract code version from gw_eccentricity.py
def get_version():
    """Get package version."""
    with open('gw_eccentricity/gw_eccentricity.py') as f:
        for line in f.readlines():
            if "__version__" in line:
                return line.split('"')[1]


setuptools.setup(
    name="gw_eccentricity",
    version=get_version(),
    author="Md Arif Shaikh, Vijay Varma, Harald Pfeiffer",
    author_email="arifshaikh.astro@gmail.com, vijay.varma392@gmail.com",
    description="Defining eccentricity for gravitational wave astronomy.",
    keywords='black-holes gravitational-waves',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vijayvarma392/gw_eccentricity",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'scipy',
        'h5py',
        'lalsuite',
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Astronomy",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Operating System :: OS Independent",
    ],
)
